#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>

#include "serial.h"

/*
   Attempt to open the serial port
   Return the file descriptor
*/
int OpenSerial(char *port)
{
   int fdes;

   if ((fdes = open(port,O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0) {
      perror("OpenSerial");
   } else {
      fcntl(fdes,F_SETFL,FNDELAY);
   }

   return(fdes);
}

/*
   Setup the communication options
   Hopefully POSIX compliant
*/
int SetupSerial(int fdes,int baud,int databits,int stopbits,int parity)
{
   int n;
   struct termios options;

   /* Get the current options */
   if (tcgetattr(fdes,&options) != 0) {
      perror("SetupSerial 1");
      return(FALSE);
   }

   /* Set the baud rate */
   switch (baud) {
   case 2400:
      n =  cfsetispeed(&options,B2400);
      n += cfsetospeed(&options,B2400);
      break;
   case 4800:
      n =  cfsetispeed(&options,B4800);
      n += cfsetospeed(&options,B4800);
      break;
   case 9600:
      n =  cfsetispeed(&options,B9600);
      n += cfsetospeed(&options,B9600);
      break;
   case 19200:
      n =  cfsetispeed(&options,B19200);
      n += cfsetospeed(&options,B19200);
      break;
   default:
      fprintf(stderr,"Unsupported baud rate\n");
      return(FALSE);
   }
   if (n != 0) {
      perror("SetupSerial 2");
      return(FALSE);
   }

   /* To check the speed, try 
   input_speed = cfgetispeed(&options);
   output_speed = cfgetospeed(&options);
   */

   /* Set the data size */
   options.c_cflag &= ~CSIZE;
   switch (databits) {
   case 7:
      options.c_cflag |= CS7;
      break;
   case 8:
      options.c_cflag |= CS8;
      break;
   default:
      fprintf(stderr,"Unsupported data size\n");
      return(FALSE);
   }

   /* Set up parity */
   switch (parity) {
   case 'n':
   case 'N':
      options.c_cflag &= ~PARENB; /* Clear parity enable */
      options.c_iflag &= ~INPCK; /* Enable parity checking */
      break;
   case 'o':
   case 'O':
      options.c_cflag |= (PARODD | PARENB); /* Enable odd parity */
      options.c_iflag |= INPCK;  /* Disnable parity checking */ 
      break;
   case 'e':
   case 'E':
      options.c_cflag |= PARENB; /* Enable parity */
      options.c_cflag &= ~PARODD; /* Turn odd off => even */
      options.c_iflag |= INPCK; /* Disnable parity checking */
      break;
   default:
      fprintf(stderr,"Unsupported parity\n");
      return(FALSE);
   }

   /* Set up stop bits */
   switch (stopbits) {
   case 1:
      options.c_cflag &= ~CSTOPB;
      break;
   case 2:
      options.c_cflag |= CSTOPB;
      break;
   default:
      fprintf(stderr,"Unsupported stop bits\n");
      return(FALSE);
   }

   /* Set input parity option */
   if (parity != 'n')
      options.c_iflag |= INPCK;

   /* Strip high order bit 
   options.c_iflag |= ISTRIP;
   */

   /* Deal with hardware or software flow control */
   options.c_cflag &= ~CRTSCTS; /* Disable RTS/CTS */
   options.c_iflag |= (IXON | IXOFF | IXANY); /* xon/xoff flow control */

   /* Output processing */
   options.c_oflag &= ~OPOST; /* No output processing */
   options.c_oflag &= ~ONLCR; /* Don't convert linefeeds */ 

   /* Input processing */
   options.c_iflag |= IGNBRK; /* Ignore break conditions */

#ifndef __DARWIN__
   options.c_iflag &= ~IUCLC; /* Don't map upper to lower case */
#endif
   options.c_iflag &= ~BRKINT; /* Ignore break signals */
   /* options.c_iflag &= ~INLCR; Map NL to CR */
   /* options.c_iflag &= ~ICRNL; Map CR to NL */

   /* Miscellaneous stuff */
   options.c_cflag |= (CLOCAL | CREAD); /* Enable receiver, set local */
	/* Linux seems to have problem with the following ??!! */
   options.c_cflag |= (IXON | IXOFF); /* Software flow control */
   options.c_lflag = 0; /* no local flags */ 
   options.c_cflag |= HUPCL; /* Drop DTR on close */
   
   /* Setup non blocking, return on 1 character */
   options.c_cc[VMIN] = 0;
   options.c_cc[VTIME] = 1;

   /* Clear the line */
   tcflush(fdes,TCIFLUSH);

   /* Update the options and do it NOW */
   if (tcsetattr(fdes,TCSANOW,&options) != 0) {
      perror("SetupSerial 3");
      return(FALSE);
   }

   return(TRUE);
}
