/* $Id$
 *
 * snark
 * Communicate with Trans-Lux Datawall displays through the Serial Port
 *
 * J. Adams (with lots of help from Mike Estee and Ramon Varra)
 * jna@retina.net
 * 11/2/2006
 * 
 * Implemented: colors, blink, justification, wipe, control codes
 * Not yet Implemented: Brightness control, flash counter, global flash, global no flash
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <termios.h>

/* Information about the display */
#define PANELID   '0'  /* override with -p */
#define MAXROWS   4
#define MAXCOLS   32
#define BRIGHTNESSROW  0x0f
#define DEVICE    "/dev/cu.USA19H3b1P1.1"

/* Prototypes */
char *formatstr(char *);
void usage(void);
void sendmessage(int fdes,int row, int column, char *s);

char panel = PANELID;

/* options */
int opts = 0; /* stdin */
int optt = 0; /* test pattern */
int optm = 0; /* messsage */
int optw = 0; /* wipe display */
int optj = 0; /* justify */
int optg = 0; /* globalflash */
int optn = 0; /* noclear */

void getstring( char *buf,  unsigned bufSize)
{
  /* safe gets() */
  unsigned dex;
  char c;
  char *cp = buf;
  
  for(dex=0; dex<bufSize-1; dex++) {
    c = getchar();
    if (c == EOF) {
      *buf = 0; return;
    }
    
    if(!isprint(c)) {
      break;
    }

    switch(c) {
    case '\n':
    case '\r':
      goto done;
    default:
      *cp++ = c;
    }
  }

 done:
  *cp = '\0';
}

void send_brightness_message(int fdes, char *s) { 
  /* I don't think my datawall supports thie feature, so i cant test it */
  char header[] = {1,panel,94};    /* <SOH><PANEL+32><BRIGHTNESS ROW><MSG> */

  /* 20h - 2Fh = brightness level.
   * 07h = blink on
   * 1Ch = blink off
   * "G" + NUMBER = flash count
   */

  /* write out the panel/row/column header */
  write(fdes, header, 3);

  /* send message out */
  write(fdes, s, strlen(s));
  write(fdes, "\004", 1);
}

void globalflash(int fdes, int option) { 
  /* flash the display */
  if (option == 1) {  /* start */
    send_brightness_message(fdes,"\007");
  }

  if (option == 2) {  /* stop */
    send_brightness_message(fdes,"\034");
  }
  
}

void wipedisplay(int fdes) {
  /* clear the entire display */
  int row;

  int oldoptj = optj;

  /* turn off justify so we can clear full lines */
  optj=0;
  for (row = 0; row < MAXROWS; row++) { 
    sendmessage(fdes,row,0,"%%C");
  }
  optj=oldoptj;

}

void sendmessage(int fdes,int row, int column, char *s) {
  /* send a message to the display */
  char str[255];

  /* TODO: Handle \n in the right way. */

  /* translux header */
  char header[] = {1,panel,0x20,0x20};    /* <SOH><PANEL><ROW><COL> */

  header[2] += row;
  strcpy(str,s);

  if (optj == 1) {
    column = 0; /* left justify */
  }

  if (optj == 2) {
    column = (MAXCOLS - strlen_nofmt(s)); /* right justify */
  }

  if (optj == 3) {
    column = ((MAXCOLS/2) - (strlen_nofmt(s)/2)) ; /* center */
  }

  header[3] += column;

  /* write out the panel/row/column header */
  write(fdes, header, 4);
  sprintf(str, formatstr(str));
  
  /* send message out */
  write(fdes, str, strlen(str));
  write(fdes, "\004", 1);
}

void usage()
{
       // print out usage
       printf("usage: snark [-r row] [-c col] (-m message | -s)\n\n" \
	      "\t-s take input from stdin (until EOF. No Horiz. or Vert. wrap)\n" \
	      "\t-n do not clear lines when reading stdin\n" \
	      "\t-m MESSAGE   display message (use -m or -s, not both.)\n\n" \
	      "\t-p N         panel ID (as a char, default '0')\n" \
	      "\t-r N         write at row N\n" \
	      "\t-c N         write at column N\n" \
	      "\t-j [l,r,c]   Justify text Left, Right, or Center\n" \
	      "\t-w           wipe display before writing\n" \
	      "\t-t           Show character set \n" \
	      "\t-g [on,off]  flash the entire panel (untested)\n" \

	      "\n\tControl characters that can be embedded in <message>:\n" \
	      "\t\t%%%%C Clear to EOL\n" \
	      "\t\t%%%%R Red text\n" \
	      "\t\t%%%%G Green text\n" \
	      "\t\t%%%%A Amber text\n" \
	      "\t\t%%%%B Start Flash\n" \
	      "\t\t%%%%X Display solid block\n" \
	      "\t\t%%%%E End Flash\n" );
}


int strlen_nofmt(char *str) { 
  /* count the length of a string, but don't count formatting */

  int len = 0;
  char *pin = str;

  /*  printf("in = %d\n",strlen(str)); */

  while (*pin != '\0') {
    if ((*pin == '%') && (*(pin+1) == '%')) { 
      pin += 3;
    }
    if (*pin >= 32) len++;
    pin++;
  }

  /*   printf("out = %d\n",len); */
  return len;
}

char *formatstr(char *str) {
  /* format a string for the translux display */
  static  char buf[1024];
  char *pin = str;
  char *pout = buf;

  while (*pin != '\0') { 
    if ((*pin == '%') && (*(pin+1) == '%')) { 
      /* take next char */
      switch (*(pin+2)) { 
      case 'X':
	*pout=127; /* squareblock */
	break;
      case 'C':
	*pout='\017'; /* clear */
	break;
      case 'R':
	*pout='\036'; /* red */
	break;
      case 'G':
	*pout='\035'; /* green */
	break;
      case'A':
	*pout='\037'; /* amber */
	break;
      case'B':
	*pout='\007'; /* blink */
	break;
      case'I':
	*pout='\032'; /* inverse - doesn't work on my datawall */
	break;
      case'E':
	*pout='\034'; /* stop flash */
	break;
      default:
	*pout='?'; /* Unknown Escape Sequence */
      }
      pin += 2;
    } else {
      *pout = *pin;
    }
    pout++;
    pin++;
  }

  *pout = '\0';

  /*  printf ("out: %s\n",buf); */
  return buf;
}

int main(int argc,char **argv)
{
   int fdes;
   int i,n,c;

   // row & column
   int row = 0;
   int column = 0;

   /* message buffer */
   char str[512];
   char input[512];

   char panelid,rowid,colid;

   struct termios newsettings,oldsettings;

   /* open the port */
   if ((fdes = OpenSerial(DEVICE)) < 0) {
      fprintf(stderr,"Port open failed\n");
      exit(-1);
   }

   /* Set port characteristics */
   if (!SetupSerial(fdes,9600,8,1,'n')) {
      fprintf(stderr,"Port setup failed\n");
      close(fdes);
      exit(-1);
   }

   /* process args */

   int ch;

   if (argc == 1) {
     printf ("You specified no parameters!\n\n");
     usage();
     exit(1);
   }

   row = 0;
   column = 0;

   while ((ch = getopt(argc, argv, "ng:p:j:whr:c:m:st")) != -1) {
     switch (ch) {
     case 't':
       /* test pattern */
       n=0;
       str[0] = '\035';

       for (i=32; i< 128;i++) { 
	 str[(n++)+1] = i;
	 if (n > MAXCOLS-1) { 
	   sendmessage(fdes,row,column,str);
	   row++;
	   n=0;
	   str[0]++;
	 }
       }

       str[n] = '\0';
       optm=1;
       optt=1;
       break;

     case 'w':
       /* don't wipe until we're configured */
       optw = 1;
       break;

     case 'g':
       if (strcmp(optarg,"on") == 0) { 
	 optg = 1;
       } 

       if (strcmp(optarg,"off") == 0) { 
	 optg = 2;
       } 
       break;

     case 's':
       opts = 1;
       break;

     case 'n':
       opts = 1;
       break;

     case 'r':
       row = atoi(optarg) - 1;
       break;

     case 'p':
       panel = optarg[0];
       break;

     case 'j': /* justify */
       switch (optarg[0]) {
       case 'l':
	 optj = 1;
	 break;
       case 'r':
	 optj = 2;
	 break;
       case 'c':
	 optj = 3;
	 break;
       }
       break;

     case 'c':
       column = atoi(optarg) - 1;
       break;

     case 'm':
       strcpy(str,optarg);
       optm = 1;
       break;

     case 'h':
       usage();
       return(0);
       break;

     default:
       /* unknown param */
       usage();
       return(1);
     }
   }

   if ((opts == 0 && optm == 0) && (optw != 1) && (optg ==0) && (optt=0)) {
     printf("You must specify one of -g, -c, -m, or -w.\n");
     usage();
   }

   if (optw) { 
     /* wipe the screen */
     wipedisplay(fdes);
   }
#ifdef EXPERIMENTAL
   if (optg > 0) { 
     globalflash(fdes,optg);
   }
#endif   
   if (opts) { 
     input[0] = 0xff; /* temp */

     /* process lines from stdin */
     while (*input != '\0') { 
       getstring(input,255);
       strcpy(str,"");

       if (optn) { 
	 strcpy(str,"\017"); 
       }
       strcat(str,input);

       /* send message */
       sendmessage(fdes,row,column,str);
       row++;
     }
   } else {
     sendmessage(fdes,row,column,str);
   }

   /* have to sleep, or it fails */
   /*   sleep(1);   */
   tcdrain(fdes);
   
   /* Close the serial port (Should restore settings) */

   /* in OSX, if I forget to close the descriptor, i don't have to sleep. This is wierd and broken. */
   /*   close(fdes); */

}



